import logging
import unittest
from batterydataextractor.nlp import AbbreviationDetector
from batterydataextractor.doc import Paragraph, Document


logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class TestAbbreviationDetection(unittest.TestCase):

    def test_abbr1(self):
        """Test the AbbreviationDetection on a simple sentence."""
        ad = AbbreviationDetector()
        self.assertEqual(
            ([('Abbr: ', 'HDAC')], [('LF: ', 'histone deacetylase')]),
            ad.detect(['as', 'histone', 'deacetylase', '(', 'HDAC', ')', 'inhibitor'])
        )

    def test_abbr2(self):
        """Test the AbbreviationDetection on a simple sentence."""
        ad = AbbreviationDetector()
        self.assertEqual(
            ([('Abbr: ', 'VPA')], [('LF: ', 'valproic acid')]),
            ad.detect(['The', 'aim', 'of', 'this', 'study', 'was', 'to', 'identify', 'valproic', 'acid', '(', 'VPA', ')'])
        )

    def test_abbr3(self):
        """Test the AbbreviationDetection on a simple sentence."""
        ad = AbbreviationDetector()
        self.assertEqual(
            ([('Abbr: ', 'THF')], [('LF: ', 'tetrahydrofuran')]),
            ad.detect(['THF', '=', 'tetrahydrofuran'])
        )

    def test_abbr4(self):
        """Test the AbbreviationDetection on a simple sentence."""
        ad = AbbreviationDetector()
        self.assertEqual(
            ([('Abbr: ', 'NAG')], [('LF: ', 'N-acetyl-β-glucosaminidase')]),
            ad.detect(['NAG', '(', 'N-acetyl-β-glucosaminidase', ')'])
        )

    def test_abbr5(self):
        """Test the AbbreviationDetection on a simple sentence."""
        ad = AbbreviationDetector()
        self.assertEqual(
            ([('Abbr: ', 'NAG')], [('LF: ', 'N-acetyl-β-glucosaminidase')]),
            ad.detect(['blood', 'urea', 'nitrogen', ',', 'N-acetyl-β-glucosaminidase', '(', 'NAG', ')', ','])
        )

    def test_equiv1(self):
        """Test the AbbreviationDetection where string equivalent is needed."""
        ad = AbbreviationDetector()
        self.assertEqual(
            ([('Abbr: ', 'CTAB')], [('LF: ', 'hexadecyltrimethylammonium bromide')]),
            ad.detect(['was', 'composed', 'of', 'hexadecyltrimethylammonium', 'bromide', '(', 'CTAB', ')', 'and'])
        )

    def test_equiv2(self):
        """Test the AbbreviationDetection where string equivalent is needed."""
        ad = AbbreviationDetector()
        self.assertEqual(
            ([('Abbr: ', 'MeOH')], [('LF: ', 'methanol')]),
            ad.detect(['was', 'mostly', 'neutral', 'in', 'methanol', '(', 'MeOH', ')'])
        )

    def test_document(self):
        elements = [
            Paragraph('''The consequences of global change on rivers include altered flow regime, and entrance of
                         compounds that may be toxic to biota. When water is scarce, a reduced dilution capacity may
                         amplify the effects of chemical pollution. Therefore, studying the response of natural
                         communities to compromised water flow and to toxicants is critical for assessing how global
                         change may affect river ecosystems. This work aims to investigate how an episode of drought
                         might influence the response of river biofilms to pulses of triclosan (TCS). The objectives
                         were to assess the separate and combined effects of simulated drought (achieved through
                         drastic flow alteration) and of TCS exposure on biofilms growing in artificial channels.'''),
        ]
        d = Document(*elements)
        self.assertEqual([[('Abbr: ', 'TCS')], [('LF: ', 'triclosan')], [('Abbr: ', 'TCS')]], d.abbreviation_definitions)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
test_reader_elsevier
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Test reader for Elsevier.
"""
import unittest
import logging
import io
import os

from batterydataextractor.doc.document import Document
from batterydataextractor.reader.elsevier import ElsevierXmlReader

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class TestElsXMLReader(unittest.TestCase):

    maxDiff = None

    def test_detect(self):
        """Test RscXMLReader can detect an RSC document."""
        r = ElsevierXmlReader()
        fname = 'els_test1.xml'
        f = io.open(os.path.join(os.path.dirname(__file__), 'testpapers', fname), 'rb')
        content = f.read()
        f.close()
        self.assertEqual(r.detect(content, fname=fname), True)

    def test_direct_usage(self):
        """Test RscXMLReader used directly to parse file."""
        r = ElsevierXmlReader()
        fname = 'els_test1.xml'
        f = io.open(os.path.join(os.path.dirname(__file__), 'testpapers', fname), 'rb')
        content = f.read()
        d = r.readstring(content)
        f.close()
        self.assertEqual(len(d.elements), 143)

    def test_document_usage(self):
        """Test RscXMLReader used via Document.from_file."""
        fname = 'els_test2.xml'
        f = io.open(os.path.join(os.path.dirname(__file__), 'testpapers', fname), 'rb')
        d = Document.from_file(f, readers=[ElsevierXmlReader()])
        self.assertEqual(len(d.elements), 56)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
batterydataextractor.reader.markup

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
XML and HTML readers based on lxml.
"""
import logging
import six
from abc import abstractmethod, ABCMeta
from collections import defaultdict

from lxml import etree
from lxml.etree import XMLParser
from lxml.html import HTMLParser

from ..errors import ReaderError
from ..doc.document import Document
from ..doc.text import Title, Heading, Paragraph, Citation, Text, Sentence, Abstract
from ..doc.meta import MetaData
from ..scrape.clean import clean
from ..scrape.csstranslator import CssHTMLTranslator
from .base import BaseReader
from ..text import get_encoding


log = logging.getLogger(__name__)

#: Inline level HTML elements
INLINE_ELEMENTS = {
    'b', 'big', 'i', 'small', 'tt', 'abbr', 'acronym', 'cite', 'code', 'dfn', 'em', 'kbd', 'strong', 'samp', 'var',
    'a', 'bdo', 'br', 'img', 'map', 'object', 'q', 'script', 'span', 'sub', 'sup', 'button', 'input', 'label',
    'select', 'textarea', 'blink', 'font', 'marquee', 'nobr', 's', 'strike', 'u', 'wbr',
}


class LxmlReader(six.with_metaclass(ABCMeta, BaseReader)):
    """Abstract base class for lxml-based readers."""

    #: A ``Cleaner`` instance to
    cleaners = [clean]

    root_css = 'html'
    title_css = 'h1'
    heading_css = 'h2, h3, h4, h5, h6'
    reference_css = 'a.ref'
    citation_css = 'cite'
    abstract_css = 'div[class="abstract"], p[class="abstract"]'

    metadata_css = 'head'
    metadata_publisher_css = 'meta[name="DC.publisher"]::attr("content"), meta[name="citation_publisher"]::attr("content")'
    metadata_author_css = 'meta[name="DC.Creator"]::attr("content"), meta[name="citation_author"]::attr("content")'
    metadata_title_css = 'meta[name="DC.title"]::attr("content"), meta[name="citation_title"]::attr("content")'
    metadata_date_css = 'meta[name="DC.Date"]::attr("content"), meta[name="citation_date"]::attr("content"), meta[name="citation_online_date"]::attr("content")'
    metadata_doi_css = 'meta[name="DC.Identifier"]::attr("content"), meta[name="citation_doi"]::attr("content")'
    metadata_language_css = 'meta[name="DC.Language"]::attr("content"), meta[name="citation_language"]::attr("content")'
    metadata_journal_css = 'meta[name="citation_journal_title"]::attr("content")'
    metadata_volume_css = 'meta[name="citation_volume"]::attr("content")'
    metadata_issue_css = 'meta[name="citation_issue"]::attr("content")'
    metadata_firstpage_css = 'meta[name="citation_firstpage"]::attr("content")'
    metadata_lastpage_css = 'meta[name="citation_lastpage"]::attr("content")'
    metadata_pdf_url_css = 'meta[name="citation_pdf_url"]::attr("content")'
    metadata_html_url_css = 'meta[name="citation_fulltext_html_url"]::attr("content"), meta[name="citation_abstract_html_url"]::attr("content")'

    ignore_css = 'a.ref sup'

    #: Inline elements
    inline_elements = INLINE_ELEMENTS

    def _parse_element_r(self, el, specials, refs, id=None, element_cls=Paragraph):
        """Recursively parse HTML/XML element and its children into a list of Document elements."""
        elements = []
        if el.tag in {etree.Comment, etree.ProcessingInstruction}:
            return []
        # if el in refs:
            # return [element_cls('', references=refs[el])]
        if el in specials:
            return specials[el]
        id = el.get('id', id)
        references = refs.get(el, [])
        if el.text is not None:
            elements.append(element_cls(six.text_type(el.text), id=id, references=references))
        elif references:
            elements.append(element_cls('', id=id, references=references))
        for child in el:
            # br is a special case - technically inline, but we want to split
            if child.tag not in {etree.Comment, etree.ProcessingInstruction} and child.tag.lower() == 'br':
                elements.append(element_cls(''))

            child_elements = self._parse_element_r(child, specials=specials, refs=refs, id=id, element_cls=element_cls)
            if (self._is_inline(child) and len(elements) > 0 and len(child_elements) > 0 and
                    isinstance(elements[-1], (Text, Sentence)) and isinstance(child_elements[0], (Text, Sentence)) and
                    type(elements[-1]) == type(child_elements[0])):
                elements[-1] += child_elements.pop(0)
            elements.extend(child_elements)
            if child.tail is not None:
                if self._is_inline(child) and len(elements) > 0 and isinstance(elements[-1], element_cls):
                    elements[-1] += element_cls(six.text_type(child.tail), id=id)
                else:
                    elements.append(element_cls(six.text_type(child.tail), id=id))
        return elements

    def _parse_element(self, el, specials=None, refs=None, element_cls=Paragraph):
        """"""
        if specials is None:
            specials = {}
        if refs is None:
            refs = {}
        elements = self._parse_element_r(el, specials=specials, refs=refs, element_cls=element_cls)
        final_elements = []
        for element in elements:
            # Filter empty text elements
            if isinstance(element, Text):
                if element.text.strip():
                    final_elements.append(element)
            else:
                final_elements.append(element)
        return final_elements

    def _parse_text(self, el, refs=None, specials=None,  element_cls=Paragraph):
        """Like _parse_element but ensure a single element."""
        if specials is None:
            specials = {}
        if refs is None:
            refs = {}
        elements = self._parse_element_r(el, specials=specials, refs=refs, element_cls=element_cls)
        # This occurs if the input element is self-closing... (some table td in NLM XML)
        if not elements:
            return [element_cls('')]
        element = elements[0]
        for next_element in elements[1:]:
            try:
                element += element_cls(' ') + next_element
            except TypeError as e:
                continue
                log.warning('Adding of two objects was skipped. {} and {} cannot be added.'.format(str(type(element)), str(type(next_element))))
        return [element]

    @staticmethod
    def _parse_reference(el):
        """Return reference ID from href or text content."""
        if '#' in el.get('href', ''):
            return [el.get('href').split('#', 1)[1]]
        elif 'rid' in el.attrib:
            return [el.attrib['rid']]
        elif 'idref' in el.attrib:
            return [el.attrib['idref']]
        else:
            return [''.join(el.itertext()).strip()]

    def _parse_metadata(self, el, refs, specials):
        title = self._css(self.metadata_title_css, el)
        authors = self._css(self.metadata_author_css, el)
        publisher = self._css(self.metadata_publisher_css, el)
        journal = self._css(self.metadata_journal_css, el)
        date = self._css(self.metadata_date_css, el)
        language = self._css(self.metadata_language_css, el)
        volume = self._css(self.metadata_volume_css, el)
        issue = self._css(self.metadata_issue_css, el)
        firstpage = self._css(self.metadata_firstpage_css, el)
        lastpage= self._css(self.metadata_lastpage_css, el)
        doi = self._css(self.metadata_doi_css, el)
        pdf_url = self._css(self.metadata_pdf_url_css, el)
        html_url = self._css(self.metadata_html_url_css, el)

        metadata = {
                '_title': title[0] if title else None,
                '_authors': authors if authors else None,
                '_publisher': publisher[0] if publisher else None,
                '_journal': journal[0] if journal else None,
                '_date': date[0] if date else None,
                '_language': language[0] if language else None,
                '_volume': volume[0] if volume else None,
                '_issue': issue[0] if issue else None,
                '_firstpage': firstpage[0] if firstpage else None,
                '_lastpage': lastpage[0] if lastpage else None,
                '_doi': doi[0] if doi else None,
                '_pdf_url': pdf_url[0] if pdf_url else None,
                '_html_url': html_url[0] if html_url else None
                }
        meta = MetaData(metadata)
        return [meta]

    @staticmethod
    def _xpath(query, root):
        result = root.xpath(query, smart_strings=False)
        if type(result) is not list:
            result = [result]
        log.debug('Selecting XPath: {}: {}'.format(query, result))
        return result

    def _css(self, query, root):
        return self._xpath(CssHTMLTranslator().css_to_xpath(query), root)

    def _is_inline(self, element):
        """Return True if an element is inline."""
        if element.tag not in {etree.Comment, etree.ProcessingInstruction} and element.tag.lower() in self.inline_elements:
            return True
        return False

    @abstractmethod
    def _make_tree(self, fstring):
        """Read a string into an lxml elementtree."""
        pass

    def parse(self, fstring):
        root = self._make_tree(fstring)
        self.root = root

        if root is None:
            raise ReaderError
        root = self._css(self.root_css, root)[0]
        for cleaner in self.cleaners:
            cleaner(root)
        specials = {}
        refs = defaultdict(list)
        titles = self._css(self.title_css, root)
        headings = self._css(self.heading_css, root)
        abstracts = self._css(self.abstract_css, root)
        citations = self._css(self.citation_css, root)
        references = self._css(self.reference_css, root)
        ignores = self._css(self.ignore_css, root)
        metadata = self._css(self.metadata_css, root)
        for reference in references:
            refs[reference.getparent()].extend(self._parse_reference(reference))
        for title in titles:
            specials[title] = self._parse_text(title, element_cls=Title, refs=refs, specials=specials)
        for heading in headings:
            specials[heading] = self._parse_text(heading, element_cls=Heading, refs=refs, specials=specials)
        for citation in citations:
            specials[citation] = self._parse_text(citation, element_cls=Citation, refs=refs, specials=specials)
        for md in metadata:
            specials[md] = self._parse_metadata(md, refs=refs, specials=specials)
        for abstract in abstracts:
            specials[abstract] = self._parse_text(abstract, element_cls=Abstract, refs=refs, specials=specials)
        for ignore in ignores:
            specials[ignore] = []
        elements = self._parse_element(root, specials=specials, refs=refs)
        return Document(*elements)


class XmlReader(LxmlReader):
    """Reader for generic XML documents."""

    def detect(self, fstring, fname=None):
        """"""
        if fname and not fname.endswith('.xml'):
            return False
        return True

    def _make_tree(self, fstring):
        root = etree.fromstring(fstring, parser=XMLParser(recover=True, encoding=get_encoding(fstring)))
        return root


class HtmlReader(LxmlReader):
    """Reader for generic HTML documents."""

    def detect(self, fstring, fname=None):
        """"""
        if fname and not (fname.endswith('.html') or fname.endswith('.htm')):
            return False
        return True

    def _make_tree(self, fstring):
        root = etree.fromstring(fstring, parser=HTMLParser(encoding=get_encoding(fstring)))
        return root



# -*- coding: utf-8 -*-
"""
test_text_normalize
~~~~~~~~~
Test the text package.
"""
import logging
import unittest

from batterydataextractor.text.normalize import normalize


logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class TestNormalization(unittest.TestCase):

    def test_normalize(self):
        """Test normalize function."""
        # Weird control characters
        self.assertEqual(u'The quick brown fox jumped', normalize(u'The\u0003 quick br\u0005own fo\u0008x jumped'))
        # Unusual whitespace characters
        self.assertEqual(u'The quick brown fox jumped', normalize(u'The\u00A0quick\u2000brown\u2008fox\u000Bjumped'))
        # u2024 instead of full stop
        self.assertEqual(u'www.bbc.co.uk', normalize(u'www\u2024bbc\u2024co\u2024uk'))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
batterydataextractor.scrape.springer

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Springer web-scraper
"""
import six
import json
import requests
from bs4 import BeautifulSoup
from .base import BaseWebScraper


class SpringerMetaWebScraper(BaseWebScraper):
    """
    Springer web-scraper using the meta API.
    """
    def __init__(self, query, api_key):
        """
        :param query: query text (e.g. battery materials)
        :param api_key: API key from the Springer API Portal (https://dev.springernature.com/)
        """
        super().__init__()
        self.api_key = api_key
        self.query = query.replace(" ", '%20')

    def get_doi(self, year=2022, start=0, max_return=100):
        """
        Get a list of dois of the query, year, and start index
        :param year: the year of published papers
        :param start: the start index of the returned papers
        :param max_return: the maximum number of returned papers (max: 100)
        :return: a list of dois of the query and year
        """
        url = "http://api.springernature.com/meta/v2/json?&q={}%20type:Journal%20year:{}&s={}&p={}&api_key={}".format(
            self.query, year, start, max_return, self.api_key)
        web_content = requests.get(url, headers={"User-Agent": "Mozilla/5.0"}).content
        content = json.loads(web_content.decode('utf-8'))
        dois = [content['records'][i]['doi'] for i in range(max_return)]
        return dois

    def get_xml(self, doi):
        """
        Get the xml link of the doi
        :param doi: the doi of a paper
        :return: the xml url
        """
        url = "https://api.springernature.com/meta/v2/pam?q=doi:{}&p=2&api_key={}".format(doi, self.api_key)
        return url

    def download_doi(self, doi, file_location):
        """
        Download the Springer xml paper
        :param doi: the doi of a paper
        :param file_location: saving location
        :return:
        """
        url = self.get_xml(doi)
        web_content = requests.get(url, headers={"User-Agent": "Mozilla/5.0"}).content
        result = self.get_spr_abstract(web_content)
        doi = result['doi'].replace("/", "_")
        date = result['date']
        name = date + "_" + doi
        with open(file_location + name + '.xml', 'wb') as f:
            f.write(web_content)
            f.close()
        return

    @staticmethod
    def get_spr_abstract(spr_document):
        """
        Get the metadata and abstract from a Springer xml file
        :param spr_document: Springer xml content
        :return: a dictionary of metadata and abstract
        """
        soup = BeautifulSoup(spr_document, features="html.parser")
        abstract = ''
        for i in soup.find_all("p"):
            if i.parent.name == "abstract":
                abstract = i.text.split("\documentclass")[0]
                abstract = abstract.replace("\n", "")
                abstract = abstract.replace("\t", "")
        date = soup.find_all("prism:publicationdate")[0].get_text()
        doi = soup.find_all("prism:doi")[0].get_text()
        journal = soup.find_all("prism:publicationname")[0].get_text()
        title = soup.find_all("dc:title")[0].get_text()
        return {"title": title, "doi": doi, "date": date, "journal": journal, "abstract": abstract}


class SpringerTDMWebScraper(BaseWebScraper):
    """
    Springer web-scraper using the TDM full-text API.
    """
    def __init__(self, query, api_key):
        """
        :param query: query text (e.g. battery materials)
        :param api_key: API key from the Springer API Portal (https://dev.springernature.com/)
        """
        super().__init__()
        self.api_key = api_key
        self.query = query.replace(" ", '%20')

    def get_doi(self, date_from="2020-06-01", date_to="2021-01-10", start=0, max_return=100):
        """
        Get a list of dois of the query, year, and start index
        :param date_from: the starting date of papers to be downloaded
        :param date_to: the end date of papers to be downloaded
        :param start: the start index of the returned papers
        :param max_return: the maximum number of returned papers (max: 100)
        :return: a list of dois of the query and year
        """
        url = "https://articles-api.springer.com/xmldata/jats?q={}%20onlinedatefrom:{}%20onlinedateto:{}&s={}" \
              "&p={}&api_key={}".format(self.query, date_from, date_to, start, max_return, self.api_key)
        web_content = requests.get(url, headers={"User-Agent": "Mozilla/5.0"}).content
        soup = BeautifulSoup(web_content, features="html.parser")
        doi = soup.find_all("article-id", attrs={"pub-id-type": "doi"})
        dois = [i.text for i in doi]
        return dois

    def get_xml(self, doi):
        """
        Get the xml link of the doi
        :param doi: the doi of a paper
        :return: the xml url
        """
        url = "https://articles-api.springer.com/xmldata/jats?q=doi:{}&api_key={}".format(doi, self.api_key)
        return url

    def download_doi(self, doi, file_location):
        """
        Download the Springer xml paper
        :param doi: the doi of a paper
        :param file_location: saving location
        :return:
        """
        url = self.get_xml(doi)
        web_content = requests.get(url, headers={"User-Agent": "Mozilla/5.0"}).content
        result = self.get_spr_abstract(web_content)
        doi = result['doi'].replace("/", "_")
        date = result['date']
        name = date + "_" + doi
        with open(file_location + name + '.xml', 'wb') as f:
            f.write(web_content)
            f.close()
        return

    @staticmethod
    def get_spr_abstract(spr_document):
        """
        Get the metadata and abstract from a Springer xml jats file
        :param spr_document: Springer xml jats content
        :return: a dictionary of metadata and abstract
        """
        soup = BeautifulSoup(spr_document, features="html.parser")
        abstract = ''
        for i in soup.find_all("p"):
            if i.parent.name == "abstract":
                abstract = i.text.split("\documentclass")[0]
                abstract = abstract.replace("\n", "")
                abstract = abstract.replace("\t", "")
        doi = soup.find_all("article-id", attrs={"pub-id-type": "doi"})[0].get_text()
        date = soup.find_all("pub-date", attrs={"date-type": "pub"})[0]
        year = date.find('year').get_text()
        try:
            month = date.find('month').get_text()
            if len(month) == 1:
                month = '0' + month
        except AttributeError:
            month = '01'
        try:
            day = date.find('day').get_text()
            if len(day) == 1:
                day = '0' + day
        except AttributeError:
            day = '01'
        date = year + month + day
        journal = soup.find_all("journal-title")[0].get_text()
        title = soup.find_all("article-title")[0].get_text()
        return {"title": title, "doi": doi, "date": date, "journal": journal, "abstract": abstract}


def spr_clean_abstract(document):
    """ Remove <title> in <abstract>"""
    # selects all tags and checks if the text or tail are spaces
    for el in document.xpath('.//abstract'):
        next_els = el.getchildren()
        for next_el in next_els:
            if next_el.tag == 'title':
                parent = next_el.getparent()
                if parent is None:
                    continue
                if next_el.tail:
                    previous = next_el.getprevious()
                    if previous is None:
                        parent.text = (parent.text or '') + next_el.tail
                    else:
                        previous.tail = (previous.tail or '') + next_el.tail
                parent.remove(next_el)
    return document


def spr_clean_ref(document):
    """Remove <article-title> in <ref>"""
    for el in document.xpath('.//ref'):
        next_els = el.getchildren()
        for next_el in next_els:
            if next_el.tag == 'mixed-citation':
                next_other_els = next_el.getchildren()
                for next_other_el in next_other_els:
                    if next_other_el.tag == 'article-title':
                        parent = next_other_el.getparent()
                        previous = next_other_el.getprevious()
                        # We can't strip the root element!
                        if parent is None:
                            continue
                        # Append the text to previous tail (or parent text if no previous), ensuring newline if block level
                        if next_other_el.text and isinstance(next_other_el.tag, six.string_types):
                            if previous is None:
                                parent.text = (parent.text or '') + next_other_el.text
                            else:
                                previous.tail = (previous.tail or '') + next_other_el.text
                        # Append the tail to last child tail, or previous tail, or parent text, ensuring newline if block level
                        if next_other_el.tail:
                            if len(next_other_el):
                                last = next_other_el[-1]
                                last.tail = (last.tail or '') + next_other_el.tail
                            elif previous is None:
                                parent.text = (parent.text or '') + next_other_el.tail
                            else:
                                previous.tail = (previous.tail or '') + next_other_el.tail
                        index = parent.index(next_other_el)
                        parent[index:index + 1] = next_other_el[:]
    return document



# -*- coding: utf-8 -*-
"""
batterydataextractor.utils

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Miscellaneous utility functions.
"""
import errno
import functools
import logging
import os

import six


log = logging.getLogger(__name__)


def memoized_property(fget):
    """Decorator to create memoized properties."""
    attr_name = '_{}'.format(fget.__name__)

    @functools.wraps(fget)
    def fget_memoized(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fget(self))
        return getattr(self, attr_name)
    return property(fget_memoized)


def memoize(obj):
    """Decorator to create memoized functions, methods or classes."""
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        if args not in cache:
            cache[args] = obj(*args, **kwargs)
        return cache[args]
    return memoizer


def python_2_unicode_compatible(klass):
    """Fix __str__, __unicode__ and __repr__ methods under Python 2."""
    if six.PY2:
        if '__str__' not in klass.__dict__:
            raise ValueError("Define __str__() on %s to use @python_2_unicode_compatible" % klass.__name__)
        if '__repr__' not in klass.__dict__:
            raise ValueError("Define __repr__() on %s to use @python_2_unicode_compatible" % klass.__name__)
        klass.__unicode__ = klass.__str__
        klass._unicode_repr = klass.__repr__
        klass.__str__ = lambda self: self.__unicode__().encode('utf-8')
        klass.__repr__ = lambda self: self._unicode_repr().encode('ascii', errors='backslashreplace')
    return klass


class Singleton(type):
    """Singleton metaclass."""
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def flatten(x):
    """Return a single flat list containing elements from nested lists."""
    result = []
    for el in x:
        if hasattr(el, '__iter__') and not isinstance(el, six.string_types):
            result.extend(flatten(el))
        else:
            result.append(el)
    return result


def first(el):
    if len(el) > 0:
        return el[0]
    else:
        return None


def ensure_dir(path):
    """Ensure a directory exists."""
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise



# -*- coding: utf-8 -*-
"""
batterydataextractor.model.base

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Data model for extracted information.
"""
import copy
from abc import ABCMeta
from collections import MutableSequence
import json
import csv
import logging

import six

log = logging.getLogger(__name__)


class BaseType(six.with_metaclass(ABCMeta)):

    # This is assigned by ModelMeta to match the attribute on the Model
    name = None

    def __init__(
            self,
            default=None,
            null=False,
            required=False,
            contextual=False,
            parse_expression=None,
            updatable=False):
        """
        :param default: (Optional) The default value for this field if none is set.
        :param bool null: (Optional) Include in serialized output even if value is None. Default False.
        :param bool required: (Optional) Whether a value is required. Default False.
        :param bool contextual: (Optional) Whether this value is contextual. Default False.
        :param BaseParserElement parse_expression: (Optional) Expression for parsing, instance of a subclass of BaseParserElement. Default None.
        :param bool updatable: (Optional) Whether the parse_expression can be changed by the document as parsing occurs. Default False
        """
        self.default = copy.deepcopy(default)
        self.null = null
        self.required = required
        self.contextual = contextual
        self.parse_expression = parse_expression
        self.updatable = updatable

        if self.parse_expression is None and self.updatable:
            print(
                'No parse_expression supplied but updatable set as True for ',
                type(self))
            print('updatable refers to whether parse_expression can be changed by the document as parsing occurs. Setting updatable to False.')
            self.updatable = False
        if updatable:
            self._default_parse_expression = copy.copy(parse_expression)

    def reset(self):
        """
        Reset the parse expression to the initial value.
        """
        if self.updatable:
            self.parse_expression = copy.copy(self._default_parse_expression)

    def __get__(self, instance, owner):
        """Descriptor for retrieving a value from a field in a Model."""
        # Check if Model class is being called, rather than Model instance
        if instance is None:
            return self
        # Get value from Model instance if available
        value = instance._values.get(self.name)
        # If value is None or empty string then return the default value, if set
        # if value in [None, ''] and self.default is not None:
        #     return self.default
        return value

    def __set__(self, instance, value):
        """Descriptor for assigning a value to a field in a Model."""
        instance._values[self.name] = self.process(value)

    def process(self, value):
        """Convert an assigned value into the desired data format for this field."""
        return value

    def serialize(self, value, primitive=False):
        """Serialize this field."""
        if hasattr(value, 'serialize'):
            # i.e. value is a nested model
            return value.serialize(primitive=primitive)
        else:
            return value


class StringType(BaseType):
    """"""

    def process(self, value):
        """Convert value to a unicode string. Useful in case lxml _ElementUnicodeResult are passed from parser."""
        return six.text_type(value) if value is not None else None


class FloatType(BaseType):
    """An floating point number field."""

    def process(self, value):
        """Convert value to a float."""
        if value is not None:
            return float(value)

        return None


class ModelType(BaseType):

    def __init__(self, model, **kwargs):
        self.model_class = model
        self.model_name = self.model_class.__name__
        super(ModelType, self).__init__(**kwargs)

    def serialize(self, value, primitive=False):
        """Serialize this field."""
        return value.serialize(primitive=primitive)


class ListType(BaseType):

    def __init__(self, field, default=None, sorted=False, **kwargs):
        super(ListType, self).__init__(**kwargs)
        self.field = field
        self.default = default if default is not None else []
        self.sorted = sorted

    def __set__(self, instance, value):
        """Descriptor for assigning a value to a ListField in a Model."""
        # Run process for the nested field type for each value in list
        if value is None:
            instance._values[self.name] = None
        else:
            processed = [self.field.process(v) for v in value]
            if self.sorted:
                processed = sorted(processed)
            instance._values[self.name] = [
                self.field.process(v) for v in value]

    def serialize(self, value, primitive=False):
        """Serialize this field."""
        return [self.field.serialize(v, primitive=primitive) for v in value]


class ModelMeta(ABCMeta):
    """"""
    def __new__(mcs, name, bases, attrs):
        cls = super(ModelMeta, mcs).__new__(mcs, name, bases, attrs)
        fields = {}
        for field_name, field in six.iteritems(cls.fields):
            fields[field_name] = copy.copy(field)
        for attr_name, attr_value in six.iteritems(attrs):
            if isinstance(attr_value, BaseType):
                # Set the name attribute on the Type to the attribute name on
                # the Model
                attr_value.name = six.text_type(attr_name)
                fields[attr_name] = attr_value
        cls.fields = fields
        parsers = []
        for parser in cls.parsers:
            p = copy.copy(parser)
            p.model = cls
            parsers.append(p)
        cls.parsers = parsers
        return cls

    def __setattr__(cls, key, value):
        if isinstance(value, BaseType):
            value.name = six.text_type(key)
            cls.fields[key] = value
        return super(ModelMeta, cls).__setattr__(key, value)

    @property
    def required_fields(cls):
        output = []
        for key, field in cls.fields.items():
            if hasattr(field, 'model_class'):
                nest_req_fields = field.model_class.required_fields
                for nrf in nest_req_fields:
                    output.append(key + '__' + nrf)
            else:
                if field.required:
                    output.append(key)
        return output


class BaseModel(six.with_metaclass(ModelMeta)):
    """"""

    fields = {}
    parsers = []
    defined_names = ['']
    specifier = None

    def __init__(self, **raw_data):
        """"""
        self._values = {}
        for key, value in six.iteritems(raw_data):
            setattr(self, key, value)
        # Set defaults
        for key, field in six.iteritems(self.fields):
            if key not in raw_data:
                setattr(self, key, copy.copy(field.default))

    @property
    def is_unidentified(self):
        """
        If there is no 'compound' field associated with the model but the compound is contextual
        """
        try:
            if 'compound' not in self.fields.keys():
                return False
            if not self.compound.contextual_fulfilled:
                return self.compound.is_unidentified
        except AttributeError:
            return True

    def __repr__(self):
        return '<%s>' % (self.__class__.__name__,)

    def __str__(self):
        return '<%s>' % (self.__class__.__name__,)

    def __eq__(self, other):
        # TODO: Check this actually works as expected (what about default
        # values?)
        if isinstance(other, self.__class__):
            log.debug(self._values, other._values)
            return self._values == other._values
        return False

    def __iter__(self):
        return iter(self.fields)

    def __delattr__(self, attr):
        """Handle deletion of field values by setting to default if specified."""
        # Set to default value
        if attr in self.fields:
            setattr(self, attr, self.fields[attr].default)
        else:
            super(BaseModel, self).__delattr__(attr)

    def __getitem__(self, key):
        """Redirect dictionary-style field access to attribute-style."""
        try:
            if key in self.fields:
                return getattr(self, key)
        except AttributeError:
            pass
        raise KeyError(key)

    def __setitem__(self, key, value):
        """Redirect dictionary-style field setting to attribute-style."""
        if key not in self.fields:
            raise KeyError(key)
        return setattr(self, key, value)

    def __contains__(self, name):
        try:
            val = getattr(self, name)
            return val is not None
        except AttributeError:
            return False

    def __hash__(self):
        return str(self.serialize()).__hash__()

    @classmethod
    def reset_updatables(cls):
        """
        Reset all updatable parse_expressions of properties associated with the class.
        """
        for key, field in six.iteritems(cls.fields):
            if cls.fields[key].updatable:
                cls.fields[key].reset()

    @classmethod
    def update(cls, definitions):
        """Update this Element's updatable attributes with new information from definitions
        Arguments:
            definitions {list} -- list of definitions found in this element
        """
        log.debug("Updating model")
        for definition in definitions:
            for field in cls.fields:
                if cls.fields[field].updatable:
                    matches = [
                        i for i in cls.fields[field].parse_expression.scan(
                            definition['tokens'])]
                    if any(matches):
                        cls.fields[field].parse_expression = cls.fields[field].parse_expression | W(
                            str(definition['specifier']))
        return

    def keys(self):
        return list(iter(self))

    def items(self):
        return [(k, getattr(self, k)) for k in self]

    def values(self):
        return [getattr(self, k) for k in self]

    def get(self, key, default=None):
        return getattr(self, key, default)

    @property
    def contextual_fulfilled(self):
        """
        Whether all the contextual fields have been extracted.
        :return: True if all fields have been found, False if not.
        :rtype: bool
        """

        for field_name, field in six.iteritems(self.fields):
            if hasattr(field, 'model_class'):
                if self[field_name] == field.default and field.contextual:
                    return False
                if hasattr(self[field_name], 'contextual_fulfilled') and \
                   not self[field_name].contextual_fulfilled:
                    log.debug('Is contextual')
                    return False
            elif field.contextual and self[field_name] == field.default:
                log.debug('Is contextual')
                return False
        log.debug('Not contextual')
        return True

    @property
    def required_fulfilled(self):
        """
        Whether all the required fields have been extracted.
        :return: True if all fields have been found, False if not.
        :rtype: bool
        """
        for field_name, field in six.iteritems(self.fields):
            if hasattr(field, 'model_class'):
                if self[field_name] == field.default \
                   and field.required:
                    return False
                if field.required and hasattr(
                        self[field_name],
                        'required_fulfilled') and not self[field_name].required_fulfilled:
                    log.debug('Required unfulfilled')
                    return False
            elif field.required and self[field_name] == field.default:
                return False
        return True

    def serialize(self, primitive=False):
        """Convert Model to python dictionary."""
        # Serialize fields to a dict
        data = {}
        for field_name in self:
            value = getattr(self, field_name)
            field = self.fields.get(field_name)
            if value is not None:
                value = field.serialize(value, primitive=primitive)
            # Skip empty fields unless field.null
            if not field.null and value in [None, '', []]:
                continue
            data[field.name] = value
        record = {self.__class__.__name__: data}
        # print(record)
        return record

    def to_json(self, *args, **kwargs):
        """Convert Model to JSON."""
        return json.dumps(self.serialize(primitive=True), *args, **kwargs)

    def to_database(self, file_name, file_type='json'):
        """
        Save the document to a file in the database.

        :param file_name: The name of the file to save the document to.
        :param file_type: The type of file to save the document to.
        """
        if file_type == 'json':
            with open(file_name, 'a', encoding='utf-8') as f:
                json.dump(self.serialize(), f)
                f.write('\n')
        elif file_type == 'txt':
            with open(file_name, 'a', encoding='utf-8') as f:
                f.write(json.dumps(self.serialize()))
                f.write('\n')
        elif file_type == 'csv':
            data = self.serialize()
            dic = data[list(data.keys())[0]]
            dic['model_type'] = list(dic.keys())[0]
            csv_columns = dic.keys()
            with open(file_name, 'a', encoding='utf-8') as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=csv_columns)
                writer.writeheader()
                writer.writerow(dic)
        else:
            raise ValueError('Unknown file type. Please use json, txt, or csv.')

    def merge_contextual(self, other):
        """
        Merges any fields marked contextual with additional information from other provided that:
        - other is of the same type and they don't have any conflicting fields
        or
        - other is a model type that is part of this model and that field is currently
        set to be the default value or the field can be merged with the other.
        .. note::
            This method mutates the model it's called on **and** returns it.
        :param other: The other model to merge into this model
        :type other: BaseModel
        :return: A merged model
        :rtype: BaseModel
        """

        log.debug(self.serialize())
        log.debug(other.serialize())
        if self.contextual_fulfilled:
            return self
        if isinstance(other, type(self)):
            # Check if the other seems to be describing the same thing as self.
            match = True
            for field_name, field in six.iteritems(self.fields):
                if (self[field_name] is not None
                    and other[field_name] is not None
                        and self[field_name] != other[field_name]):
                    match = False
                    break
            if match:
                for field_name, field in six.iteritems(self.fields):
                    if (field.contextual
                        and self[field_name] is None
                            and other.get(field_name, None) is not None):
                        self[field_name] = other[field_name]
        else:
            for field_name, field in six.iteritems(self.fields):
                if hasattr(
                        field,
                        'model_class') and isinstance(
                        other,
                        field.model_class):
                    log.debug('model class case')
                    if self[field_name] is not None and not self[field_name].contextual_fulfilled:
                        self[field_name] = self[field_name].merge_contextual(
                            other)
                    elif field.contextual and self[field_name] is None:
                        log.debug(field_name)
                        self[field_name] = copy.copy(other)
        return self

    def merge_all(self, other):
        """
        Merges any properties between other and self, regardless of whether that field is contextual.
        Checks to make sure that there are no conflicts between the values contained in self and those in other.
        .. note::
            This method mutates the model it's called on **and** returns it.
        :param other: The other model to merge into this model
        :type other: BaseModel
        :return: A merged model
        :rtype: BaseModel
        """

        log.debug(self.serialize())
        log.debug(other.serialize())
        if isinstance(other, type(self)):
            # Check if the other seems to be describing the same thing as self.
            match = True
            for field_name, field in six.iteritems(self.fields):
                if (self[field_name] is not None
                    and other[field_name] is not None
                        and self[field_name] != other[field_name]):
                    match = False
                    break
            if match:
                for field_name, field in six.iteritems(self.fields):
                    if (self[field_name] is None
                            and other.get(field_name, None) is not None):
                        self[field_name] = other[field_name]
        else:
            for field_name, field in six.iteritems(self.fields):
                if hasattr(
                        field,
                        'model_class') and isinstance(
                        other,
                        field.model_class):
                    log.debug('model class case')
                    if self[field_name] is not None:
                        self[field_name] = self[field_name].merge_all(
                            other)
                    elif self[field_name] is None:
                        log.debug(field_name)
                        self[field_name] = copy.copy(other)
        return self


class ModelList(MutableSequence):
    """Wrapper around a list of Models objects to facilitate operations on all at once."""

    def __init__(self, *models):
        self.models = list(models)

    def __getitem__(self, index):
        return self.models[index]

    def __setitem__(self, index, value):
        self.models[index] = value

    def __delitem__(self, index):
        del self.models[index]

    def __len__(self):
        return len(self.models)

    def __repr__(self):
        return self.models.__repr__()

    def __str__(self):
        return self.models.__str__()

    def __contains__(self, element):
        log.debug(element.serialize())
        log.debug(self.serialize())
        log.debug(self.models.__contains__(element))
        return self.models.__contains__(element)

    def insert(self, index, value):
        self.models.insert(index, value)

    def serialize(self):
        """Serialize to a list of python dictionaries."""
        return [e.serialize() for e in self.models]

    def to_json(self, *args, **kwargs):
        """Convert ModelList to JSON."""
        return json.dumps(self.serialize(), *args, **kwargs)



# -*- coding: utf-8 -*-
"""
batterydataextractor.nlp.lexicon

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Cache features of previously seen words. (No Brown word clusters)
When processing documents, BDE adds each unique word that it encounters to the Lexicon as a Lexeme.
Each Lexeme stores various word features, so they don’t have to be re-calculated for every occurrence of that word.
"""
import logging

import six

from ..text import word_shape, is_ascii, is_punct, like_url, like_number
from ..text.normalize import Normalizer, ChemNormalizer
from ..utils import Singleton

log = logging.getLogger(__name__)


class Lexeme(object):
    """"""

    __slots__ = ('text', 'normalized', 'lower', 'first', 'suffix', 'shape', 'length', 'upper_count', 'lower_count',
                 'digit_count', 'is_alpha', 'is_ascii', 'is_digit', 'is_lower', 'is_upper', 'is_title', 'is_punct',
                 'is_hyphenated', 'like_url', 'like_number', 'cluster')

    def __init__(self, text, normalized, lower, first, suffix, shape, length, upper_count, lower_count, digit_count,
                 is_alpha, is_ascii, is_digit, is_lower, is_upper, is_title, is_punct, is_hyphenated, like_url,
                 like_number):
        #: Original Lexeme text.
        self.text = text
        #: Normalized text, using the Lexicon Normalizer.
        self.normalized = normalized
        #: Lowercase text.
        self.lower = lower
        #: First character.
        self.first = first
        #: Three-character suffix
        self.suffix = suffix
        #: Word shape. Derived by replacing every number with 'd', every greek letter with 'g', and every latin letter with 'X' or 'x' for uppercase and lowercase respectively.
        self.shape = shape
        #: Lexeme length.
        self.length = length
        #: Count of uppercase characters.
        self.upper_count = upper_count
        #: Count of lowercase characters.
        self.lower_count = lower_count
        #: Count of digits.
        self.digit_count = digit_count
        #: Whether the text is entirely alphabetical characters.
        self.is_alpha = is_alpha
        #: Whether the text is entirely ASCII characters.
        self.is_ascii = is_ascii
        #: Whether the text is entirely digits.
        self.is_digit = is_digit
        #: Whether the text is entirely lowercase.
        self.is_lower = is_lower
        #: Whether the text is entirely uppercase.
        self.is_upper = is_upper
        #: Whether the text is title cased.
        self.is_title = is_title
        #: Whether the text is entirely punctuation characters.
        self.is_punct = is_punct
        #: Whether the text is hyphenated.
        self.is_hyphenated = is_hyphenated
        #: Whether the text looks like a URL.
        self.like_url = like_url
        #: Whether the text looks like a number.
        self.like_number = like_number


class Lexicon(six.with_metaclass(Singleton)):
    """"""

    #: The Normalizer for this Lexicon.
    normalizer = Normalizer()

    def __init__(self):
        """"""
        self.lexemes = {}

    def __len__(self):
        """The current number of lexemes stored."""
        return len(self.lexemes)

    def add(self, text):
        """Add text to the lexicon.
        :param string text: The text to add.
        """
        # logging.debug('Adding to lexicon: %s' % text)
        if text not in self.lexemes:
            normalized = self.normalized(text)
            self.lexemes[text] = Lexeme(
                text=text,
                normalized=normalized,
                lower=self.lower(normalized),
                first=self.first(normalized),
                suffix=self.suffix(normalized),
                shape=self.shape(normalized),
                length=self.length(normalized),
                upper_count=self.upper_count(normalized),
                lower_count=self.lower_count(normalized),
                digit_count=self.digit_count(normalized),
                is_alpha=self.is_alpha(normalized),
                is_ascii=self.is_ascii(normalized),
                is_digit=self.is_digit(normalized),
                is_lower=self.is_lower(normalized),
                is_upper=self.is_upper(normalized),
                is_title=self.is_title(normalized),
                is_punct=self.is_punct(normalized),
                is_hyphenated=self.is_hyphenated(normalized),
                like_url=self.like_url(normalized),
                like_number=self.like_number(normalized),
            )

    def __getitem__(self, text):
        """Return the requested lexeme from the Lexicon.
        :param string text: Text of the lexeme to retrieve.
        :rtype: Lexeme
        :returns: The requested Lexeme.
        """
        self.add(text)
        return self.lexemes[text]

    def normalized(self, text):
        """"""
        return self.normalizer(text)

    @staticmethod
    def lower(text):
        """"""
        return text.lower()

    @staticmethod
    def first(text):
        """"""
        return text[:1]

    @staticmethod
    def suffix(text):
        """"""
        return text[-3:]

    @staticmethod
    def shape(text):
        """"""
        return word_shape(text)

    @staticmethod
    def length(text):
        """"""
        return len(text)

    @staticmethod
    def digit_count(text):
        """"""
        return sum(c.isdigit() for c in text)

    @staticmethod
    def upper_count(text):
        """"""
        return sum(c.isupper() for c in text)

    @staticmethod
    def lower_count(text):
        """"""
        return sum(c.islower() for c in text)

    @staticmethod
    def is_alpha(text):
        """"""
        return text.isalpha()

    @staticmethod
    def is_ascii(text):
        """"""
        return is_ascii(text)

    @staticmethod
    def is_digit(text):
        """"""
        return text.isdigit()

    @staticmethod
    def is_lower(text):
        """"""
        return text.islower()

    @staticmethod
    def is_upper(text):
        """"""
        return text.isupper()

    @staticmethod
    def is_title(text):
        """"""
        return text.istitle()

    @staticmethod
    def is_punct(text):
        """"""
        return is_punct(text)

    @staticmethod
    def is_hyphenated(text):
        """"""
        return '-' in text and not text == '-'

    @staticmethod
    def like_url(text):
        """"""
        return like_url(text)

    @staticmethod
    def like_number(text):
        """"""
        return like_number(text)


class ChemLexicon(Lexicon):
    """A Lexicon that is pre-configured with a Chemistry-aware Normalizer and Brown word clusters derived from a
    chemistry corpus."""

    normalizer = ChemNormalizer()



# -*- coding: utf-8 -*-
"""
batterydataextractor.reader.plaintext

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Plain text document reader.
"""
import re
import six

from ..doc.document import Document
from .base import BaseReader
from ..text import get_encoding


class PlainTextReader(BaseReader):
    """Read plain text and split into Paragraphs based on newline patterns."""

    def detect(self, fstring, fname=None):
        """Have a stab at most files."""
        if fname is not None and '.' in fname:
            extension = fname.rsplit('.', 1)[1]
            if extension in {'pdf', 'html', 'xml'}:
                return False
        return True

    def parse(self, fstring):
        if isinstance(fstring, six.binary_type):
            fstring = fstring.decode(get_encoding(fstring))
        para_strings = [p.strip() for p in re.split(r'\r\n[ \t]*\r\n|\r[ \t]*\r|\n[ \t]*\n', fstring)]
        return Document(*para_strings)



# -*- coding: utf-8 -*-
"""
Extend cssselect to improve handling of pseudo-elements.
This is derived from csstranslator.py in the Scrapy project. The original file is available at:
https://github.com/scrapy/scrapy/blob/master/scrapy/selector/csstranslator.py
The original file was released under the BSD license:
Copyright (c) Scrapy developers.
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
    3. Neither the name of Scrapy nor the names of its contributors may be used
       to endorse or promote products derived from this software without
       specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from cssselect import GenericTranslator, HTMLTranslator
from cssselect.xpath import _unicode_safe_getattr, XPathExpr, ExpressionError
from cssselect.parser import FunctionalPseudoElement


class CdeXPathExpr(XPathExpr):

    textnode = False
    attribute = None

    @classmethod
    def from_xpath(cls, xpath, textnode=False, attribute=None):
        x = cls(path=xpath.path, element=xpath.element, condition=xpath.condition)
        x.textnode = textnode
        x.attribute = attribute
        return x

    def __str__(self):
        path = super(CdeXPathExpr, self).__str__()
        if self.textnode:
            if path == '*':
                path = 'text()'
            elif path.endswith('::*/*'):
                path = path[:-3] + 'text()'
            else:
                path += '/text()'
        if self.attribute is not None:
            if path.endswith('::*/*'):
                path = path[:-2]
            path += '/@%s' % self.attribute
        return path

    def join(self, combiner, other):
        super(CdeXPathExpr, self).join(combiner, other)
        self.textnode = other.textnode
        self.attribute = other.attribute
        return self


class TranslatorMixin(object):

    def xpath_element(self, selector):
        xpath = super(TranslatorMixin, self).xpath_element(selector)
        return CdeXPathExpr.from_xpath(xpath)

    def xpath_pseudo_element(self, xpath, pseudo_element):
        if isinstance(pseudo_element, FunctionalPseudoElement):
            method = 'xpath_%s_functional_pseudo_element' % (pseudo_element.name.replace('-', '_'))
            method = _unicode_safe_getattr(self, method, None)
            if not method:
                raise ExpressionError("The functional pseudo-element ::%s() is unknown" % pseudo_element.name)
            xpath = method(xpath, pseudo_element)
        else:
            method = 'xpath_%s_simple_pseudo_element' % (pseudo_element.replace('-', '_'))
            method = _unicode_safe_getattr(self, method, None)
            if not method:
                raise ExpressionError("The pseudo-element ::%s is unknown" % pseudo_element)
            xpath = method(xpath)
        return xpath

    def xpath_attr_functional_pseudo_element(self, xpath, function):
        if function.argument_types() not in (['STRING'], ['IDENT']):
            raise ExpressionError("Expected a single string or ident for ::attr(), got %r" % function.arguments)
        return CdeXPathExpr.from_xpath(xpath, attribute=function.arguments[0].value)

    def xpath_text_simple_pseudo_element(self, xpath):
        """Support selecting text nodes using ::text pseudo-element"""
        return CdeXPathExpr.from_xpath(xpath, textnode=True)


class CssXmlTranslator(TranslatorMixin, GenericTranslator):
    pass


class CssHTMLTranslator(TranslatorMixin, HTMLTranslator):
    pass



# -*- coding: utf-8 -*-
"""
batterydataextractor.parse.cem

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Chemical entity mention parser elements.
"""
import logging
from .base import BaseSentenceParser
from ..nlp import CemTagger

log = logging.getLogger(__name__)


class CompoundParser(BaseSentenceParser):
    """Chemical name possibly with an associated label."""

    ct = CemTagger()

    def interpret(self, tokens):
        cems = self.ct.tag(tokens)
        for cem in cems:
            if cem[-1] == 'MAT':
                c = self.model(names=[cem[0]])
                yield c



# -*- coding: utf-8 -*-
"""
batterydataextractor.reader.base

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Base reader
"""
from abc import ABCMeta, abstractmethod
import six


class BaseReader(six.with_metaclass(ABCMeta)):
    """All Document Readers should implement a parse method."""

    def __init__(self):
        self.root = None

    def detect(self, fstring, fname=None):
        """Quickly check if this reader can parse the input. Reader subclasses should override this.
        Used to quickly skip attempting to parse when trying different readers. If in doubt, return True, and then
        raise ReaderError in the parse method if it fails.
        """
        return True

    @abstractmethod
    def parse(self, fstring):
        """Parse the input and return a Document. Raises ReaderError if the parse fails."""
        pass

    def read(self, f):
        """Read a file-like object and return a Document."""
        return self.parse(f.read())

    def readstring(self, fstring):
        """Read a file string and return a Document."""
        return self.parse(fstring)



# -*- coding: utf-8 -*-
"""
batterydataextractor.scrape.elsevier

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Elsevier web-scraper
"""
import six
import json
import requests
import numpy as np
from bs4 import BeautifulSoup
from .base import BaseWebScraper


class ElsevierWebScraper(BaseWebScraper):
    """
    Elsevier web-scraper
    """
    base_url = 'https://api.elsevier.com/content/search/sciencedirect'

    def __init__(self, api_key, query):
        """
        :param api_key: API key from the Elsevier Developer Portal (https://dev.elsevier.com/)
        :param query: query text (e.g. battery materials)
        """
        super().__init__()
        self.api_key = api_key
        self.data = {"qs": query,
                     "date": 2022,
                     # "volume": 0,
                     "display": {"show": 100, "offset": 0}
                     }
        self.headers = {'x-els-apikey': api_key, 'Content-Type': 'application/json', 'Accept': 'application/json'}

    def get_response(self):
        """
        Get the requests response from the url.
        :return: requests response.
        """
        response = requests.put(self.base_url, data=json.dumps(self.data), headers=self.headers)
        response = response.text.replace('false', 'False').replace('true', 'True')
        try:
            response = eval(response)
        except BaseException:
            print(response)
        return response

    def get_doi(self, year=2022, volume=None):
        """
        Get the list of dois from the query text, year, and volume.
        :param year: the year of published papers
        :param volume: the volume of the journal
        :return:
        """
        dois = []
        if volume is not None:
            self.data['volume'] = volume
        self.data["date"] = year
        response = self.get_response()
        if 'resultsFound' in response.keys():
            n = int(np.ceil(response['resultsFound'] / 100))
        else:
            n = 60
        for offset in range(1 + n):
            self.data["display"]["offset"] = offset
            response = self.get_response()
            if 'results' in response.keys():
                results = response['results']
                for result in results:
                    if 'doi' in result:
                        dois.append(result['doi'])
        return list(set(dois))

    def download_doi(self, doi, file_location):
        """
        Download the xml file from the doi
        :param doi: doi of a paper
        :param file_location: saving location
        :return:
        """
        request_url = 'https://api.elsevier.com/content/article/doi/{}?apiKey={}&httpAccept=text%2Fxml'.format(
            doi, self.api_key)
        text = requests.get(request_url).text
        result = self.get_els_abstract(text)
        doi = result['doi'].replace("/", "_")
        date = result['date'].replace("-", "")
        name = date + "_" + doi
        with open(file_location + name + '.xml', 'w', encoding='utf-8') as f:
            f.write(text)
        return

    @staticmethod
    def get_els_abstract(els_document):
        """
        Get the metadata and abstract from a Elsevier xml file
        :param els_document: Elsevier xml content
        :return: a dictionary of metadata and abstract
        """
        soup = BeautifulSoup(els_document, features="html.parser")
        date = soup.find_all("xocs:available-online-date")[0].get_text()
        title = soup.find_all("dc:title")[0].get_text()
        journal = soup.find_all("prism:publicationname")[0].get_text()
        abstract = soup.find_all("dc:description")[0].get_text()
        abstract = abstract.replace("\n", "")
        abstract = abstract.replace("\t", "")
        doi = soup.find_all("prism:doi")[0].get_text()
        return {"title": title, "doi": doi, "date": date, "journal": journal, "abstract": abstract}


def fix_elsevier_xml_whitespace(document):
    """ Fix tricky xml tags"""
    # space hsp  and refs correctly
    for el in document.xpath('.//ce:hsp'):
        parent = el.getparent()
        previous = el.getprevious()
        if parent is None:
            continue
        # Append the text to previous tail (or parent text if no previous), ensuring newline if block level
        if el.text and isinstance(el.tag, six.string_types):
            if previous is None:
                if parent.text:
                    if parent.text.endswith(' '):
                        parent.text = (parent.text or '') + '' + el.text
                    else:
                        parent.text = (parent.text or '') + ' ' + el.text
            else:
                if previous.tail:
                    if previous.tail.endswith(' '):
                        previous.tail = (previous.tail or '') + '' + el.text
                    else:
                        previous.tail = (previous.tail or '') + ' ' + el.text
        # Append the tail to last child tail, or previous tail, or parent text, ensuring newline if block level
        if el.tail:
            if len(el):
                last = el[-1]
                last.tail = (last.tail or '') + el.tail
            elif previous is None:
                if el.tail.startswith(' '):
                    parent.text = (parent.text or '') + '' + el.tail
                else:
                    parent.text = (parent.text or '') + ' ' + el.tail
            else:
                if el.tail.startswith(' '):
                    previous.tail = (previous.tail or '') + '' + el.tail
                else:
                    previous.tail = (previous.tail or '') + ' ' + el.tail

        index = parent.index(el)
        parent[index:index + 1] = el[:]
    return document


def els_xml_whitespace(document):
    """ Remove whitespace in xml.text or xml.tails for all elements, if it is only whitespace """
    # selects all tags and checks if the text or tail are spaces
    for el in document.xpath('//*'):
        if str(el.text).isspace():
            el.text = ''
        if str(el.tail).isspace():
            el.tail = ''
        if el.text:
            el.text = el.text.replace('\n', ' ')
    return document


def els_clean_abstract(document):
    """ Remove <ce:section-title> in <ce:abstract>"""
    # selects all tags and checks if the text or tail are spaces
    for el in document.xpath('.//ce:abstract'):
        next_els = el.getchildren()
        for next_el in next_els:
            if next_el.tag == '{http://www.elsevier.com/xml/common/dtd}section-title':
                parent = next_el.getparent()
                if parent is None:
                    continue
                if next_el.tail:
                    previous = next_el.getprevious()
                    if previous is None:
                        parent.text = (parent.text or '') + next_el.tail
                    else:
                        previous.tail = (previous.tail or '') + next_el.tail
                parent.remove(next_el)
    return document



import logging
import unittest
from batterydataextractor.doc import Document

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class TestPropertyDataMp(unittest.TestCase):

    def do_parse(self, input, expected):
        p = Document(input)
        p.add_models_by_names(["mp"])
        log.debug(p)
        log.debug([r.serialize() for r in p.records])
        self.assertEqual(expected, [r.serialize() for r in p.records])

    def test_mpc1(self):
        s = 'NaCl: mp 163-164 °C.'
        expected = [{'PropertyData': {'confidence_score': 0.608,
                                      'material': 'NaCl',
                                      'raw_value': '163-164 ° C',
                                      'specifier': 'mp',
                                      'units': '° C',
                                      'value': [163.0, 164.0]}}]
        self.do_parse(s, expected)

    def test_mpc2(self):
        s = 'Bromo-2,6-dichloroaniline: mp 71-72° C.'
        expected = [{'PropertyData': {'confidence_score': 0.9909,
                                      'material': 'Bromo-2,6-dichloroaniline',
                                      'raw_value': '71-72 ° C',
                                      'specifier': 'mp',
                                      'units': '° C',
                                      'value': [71.0, 72.0]}}]
        self.do_parse(s, expected)


class TestOriginalText(unittest.TestCase):

    def do_parse(self, input, expected):
        p = Document(input)
        p.add_models_by_names(["capacity", "voltage"], original_text=True)
        log.debug(p)
        log.debug([r.serialize() for r in p.records])
        self.assertEqual(expected, [r.serialize() for r in p.records])

    def test_original_test(self):
        s = "The theoretical capacity of graphite is 372 mAh/g... In the case of LiFePO4 chemistry, the absolute " \
            "maximum voltage is 4.2V per cell."
        expected = [{'PropertyData': {'confidence_score': 0.6054, 'material': 'graphite',
                                      'original_text': 'The theoretical capacity of graphite is 372 mAh / g ... In the '
                                                       'case of LiFePO4 chemistry , the absolute maximum voltage is '
                                                       '4.2 V per cell .',
                                      'raw_value': '372 mAh / g', 'specifier': 'capacity', 'units': 'mAh / g',
                                      'value': [372.0]}},
                    {'PropertyData': {'confidence_score': 0.7035, 'material': 'LiFePO4',
                                      'original_text': 'The theoretical capacity of graphite is 372 mAh / g ... '
                                                       'In the case of LiFePO4 chemistry , the absolute maximum '
                                                       'voltage is 4.2 V per cell .',
                                      'raw_value': '4.2 V', 'specifier': 'voltage', 'units': 'V', 'value': [4.2]}}]
        self.do_parse(s, expected)


class TestGeneralInfoApparatus(unittest.TestCase):

    def do_parse(self, input, expected):
        p = Document(input)
        p.add_general_models(["apparatus"])
        log.debug(p)
        log.debug([r.serialize() for r in p.records])
        self.assertEqual(expected, [r.serialize() for r in p.records])

    def test_appra1(self):
        s = '1H NMR spectra were recorded on a Varian MR-400 MHz instrument.'
        expected = [{'GeneralInfo':
                         {'answer': 'Varian MR - 400 MHz instrument',
                          'confidence_score': 0.5065,
                          'specifier': 'apparatus'}}]
        self.do_parse(s, expected)

    def test_appra2(self):
        s = 'The photoluminescence quantum yield (PLQY) was measured using a FluoroMax-4 spectrofluorimeter.'
        expected = [{'GeneralInfo':
                                {'answer': 'FluoroMax-4 spectrofluorimeter',
                                 'confidence_score': 0.8481,
                                 'specifier': 'apparatus'}}]
        self.do_parse(s, expected)


class TestGeneralInfoElectrolyte(unittest.TestCase):

    def do_parse(self, input, expected):
        p = Document(input)
        p.add_general_models(["electrolyte"])
        log.debug(p)
        log.debug([r.serialize() for r in p.records])
        self.assertEqual(expected, [r.serialize() for r in p.records])

    def test_electrolyte(self):
        s = 'The typical non-aqueous electrolyte for commercial Li-ion cells is a solution of LiPF6 in linear and ' \
            'cyclic carbonates such as dimethyl carbonate and ethylene carbonate, respectively [1], [2].'
        expected = [{'GeneralInfo': {'answer': 'a solution of LiPF6', 'confidence_score': 0.3122,
                                     'specifier': 'electrolyte'}}]
        self.do_parse(s, expected)


class TestGeneralInfoDeviceComponent(unittest.TestCase):

    def do_parse(self, input, expected):
        p = Document(input)
        p.add_general_models(["anode", "cathode"])
        log.debug(p)
        log.debug([r.serialize() for r in p.records])
        self.assertEqual(expected, [r.serialize() for r in p.records])

    def test_device_component(self):
        s = 'The lithium iron phosphate battery (LiFePO4 battery) or LFP battery (lithium ferrophosphate), is a type ' \
            'of lithium-ion battery using lithium iron phosphate (LiFePO4) as the cathode material, and a graphitic ' \
            'carbon electrode with a metallic backing as the anode.'
        expected = [{'GeneralInfo': {'answer': 'graphitic carbon', 'confidence_score': 0.7898, 'specifier': 'anode'}},
                    {'GeneralInfo': {'answer': 'lithium iron phosphate', 'confidence_score': 0.9312,
                                     'specifier': 'cathode'}}]
        self.do_parse(s, expected)


if __name__ == '__main__':
    unittest.main()



import unittest
from batterydataextractor.scrape import RSCWebScraper


class TestRSCScraper(unittest.TestCase):

    rsc_scraper = RSCWebScraper()

    def test_get_doi_rsc(self):
        """
        Test if the list of dois can be found with the RSCWebScraper
        """
        dois = self.rsc_scraper.get_doi(query="battery materials", page=3)
        length = len(dois)

        self.assertEqual(length, 25)

    def test_els_scraper(self):
        # As Elsevier web-scraper requires API key, we have no tests here.
        # See `examples/webscrape/elsevier.py` for more details.
        pass

    def test_spr_scraper(self):
        # As Springer web-scraper requires API key, we have no tests here.
        # See `examples/webscrape/spr.py` for more details.
        pass


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
batterydataextractor.doc.document

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Document model.
"""
from abc import ABCMeta, abstractmethod
import collections
import io
import json
import csv
import logging

import six

from .text import Paragraph, Citation, Footnote, Heading, Title, Caption
from .element import CaptionedElement
from .meta import MetaData
from ..errors import ReaderError
from ..model.base import ModelList
from ..model.model import PropertyData, Compound, GeneralInfo
from ..text import get_encoding
from ..config import Config


log = logging.getLogger(__name__)


class BaseDocument(six.with_metaclass(ABCMeta, collections.Sequence)):
    """Abstract base class for a Document."""

    def __repr__(self):
        return '<%s: %s elements>' % (self.__class__.__name__, len(self))

    def __str__(self):
        return '<%s: %s elements>' % (self.__class__.__name__, len(self))

    def __getitem__(self, index):
        return self.elements[index]

    def __len__(self):
        return len(self.elements)

    @property
    @abstractmethod
    def elements(self):
        """Return a list of document elements."""
        return []

    @property
    @abstractmethod
    def records(self):
        """Chemical records that have been parsed from this Document."""
        return []


class Document(BaseDocument):
    """A document to extract data from. Contains a list of document elements."""
    # TODO: Add a usage example here in the documentation.

    def __init__(self, *elements, **kwargs):
        """Initialize a Document manually by passing one or more Document elements (Paragraph, Heading, Table, etc.)
        Strings that are passed to this constructor are automatically wrapped into Paragraph elements.
        :param list[batterydataextractor.doc.element.BaseElement|string] elements: Elements in this Document.
        :keyword Config config: (Optional) Config file for the Document.
        :keyword list[BaseModel] models: (Optional) Models that the Document should extract data for.
        """
        self._elements = []
        for element in elements:
            # Convert raw text to Paragraph elements
            if isinstance(element, six.text_type):
                element = Paragraph(element)
            elif isinstance(element, six.binary_type):
                # Try guess encoding if byte string
                encoding = get_encoding(element)
                log.warning('Guessed bytestring encoding as %s. Use unicode strings to avoid this warning.', encoding)
                element = Paragraph(element.decode(encoding))
            element.document = self
            self._elements.append(element)
        if 'config' in kwargs.keys():
            self.config = kwargs['config']
        else:
            self.config = Config()
        if 'models' in kwargs.keys():
            self._models = kwargs['models']
        else:
            self._models = []

        if 'device' in kwargs.keys():
            self._device = kwargs['device']
        else:
            self._device = -1

        # Sets parameters from configuration file
        for element in elements:
            if callable(getattr(element, 'set_config', None)):
                element.set_config()
        log.debug('%s: Initializing with %s elements' % (self.__class__.__name__, len(self.elements)))

    def add_models(self, models):
        """
        Add models to all elements.
        Usage::
            d = Document.from_file(f)
            d.add_models([myModelClass1, myModelClass2,..])
        Arguments::
            models -- List of model classes
        """
        log.debug("Setting models")
        self._models.extend(models)
        for element in self.elements:
            if callable(getattr(element, 'add_models', None)):
                element.add_models(models)
        return

    def add_models_by_names(self, names, confidence_threshold=0, original_text=False):
        """
        Add models to all elements.
        Usage::
            d = Document.from_file(f)
            d.add_models_by_names("myModelName1", "myModelName2",..])
        Arguments::
            models -- List of model classes
        """
        log.debug("Setting models by names")

        model = PropertyData
        model.defined_names = names
        model.confidence_threshold = confidence_threshold
        model.original_text = original_text
        model.device = self.device
        self._models.extend([model])
        for element in self.elements:
            if callable(getattr(element, 'add_models', None)):
                element.add_models([model])
        return

    def add_general_models(self, names, confidence_threshold=0, original_text=False, self_defined=False):
        """
        Add models to all elements.
        Usage::
            d = Document.from_file(f)
            d.add_general_models("myModelName1", "myModelName2",..])
        Arguments::
            models -- List of model classes
        """
        log.debug("Setting models by names")

        model = GeneralInfo
        model.defined_names = names
        model.confidence_threshold = confidence_threshold
        model.original_text = original_text
        model.self_defined = self_defined
        model.device = self.device
        self._models.extend([model])
        for element in self.elements:
            if callable(getattr(element, 'add_models', None)):
                element.add_models([model])
        return

    @property
    def models(self):
        return self._models

    @models.setter
    def models(self, value):
        self._models = value
        for element in self.elements:
            element.models = value

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, value):
        self._device = value
        for element in self.elements:
            element.device = value

    @classmethod
    def from_file(cls, f, fname=None, readers=None):
        """Create a Document from a file.
        Usage::
            with open('paper.html', 'rb') as f:
                doc = Document.from_file(f)
        .. note::
            Always open files in binary mode by using the 'rb' parameter.
        :param f: A file-like object or path to a file.
        :type f: file or str
        :param str fname: (Optional) The filename. Used to help determine file format.
        :param list[batterydataextractor.reader.base.BaseReader] readers: (Optional) List of readers to use. If not set, Document will try all default readers,
            which are :class:`~batterydataextractor.reader.acs.AcsHtmlReader`, :class:`~batterydataextractor.reader.rsc.RscHtmlReader`,
            :class:`~batterydataextractor.reader.nlm.NlmXmlReader`, :class:`~batterydataextractor.reader.uspto.UsptoXmlReader`,
            :class:`~batterydataextractor.reader.cssp.CsspHtmlReader`, :class:`~batterydataextractor.elsevier.ElsevierXmlReader`,
            :class:`~batterydataextractor.reader.markup.XmlReader`, :class:`~batterydataextractor.reader.markup.HtmlReader`,
            :class:`~batterydataextractor.reader.pdf.PdfReader`, and :class:`~batterydataextractor.reader.plaintext.PlainTextReader`.
        """
        if isinstance(f, six.string_types):
            f = io.open(f, 'rb')
        if not fname and hasattr(f, 'name'):
            fname = f.name
        return cls.from_string(f.read(), fname=fname, readers=readers)

    @classmethod
    def from_string(cls, fstring, fname=None, readers=None):
        """Create a Document from a byte string containing the contents of a file.
        Usage::
            contents = open('paper.html', 'rb').read()
            doc = Document.from_string(contents)
        .. note::
            This method expects a byte string, not a unicode string (in contrast to most methods in batterydataextractor).
        :param bytes fstring: A byte string containing the contents of a file.
        :param str fname: (Optional) The filename. Used to help determine file format.
        :param list[batterydataextractor.reader.base.BaseReader] readers: (Optional) List of readers to use. If not set, Document will try all default readers,
            which are :class:`~batterydataextractor.reader.acs.AcsHtmlReader`, :class:`~batterydataextractor.reader.rsc.RscHtmlReader`,
            :class:`~batterydataextractor.reader.nlm.NlmXmlReader`, :class:`~batterydataextractor.reader.uspto.UsptoXmlReader`,
            :class:`~batterydataextractor.reader.cssp.CsspHtmlReader`, :class:`~batterydataextractor.elsevier.ElsevierXmlReader`,
            :class:`~batterydataextractor.reader.markup.XmlReader`, :class:`~batterydataextractor.reader.markup.HtmlReader`,
            :class:`~batterydataextractor.reader.pdf.PdfReader`, and :class:`~batterydataextractor.reader.plaintext.PlainTextReader`.
        """
        if readers is None:
            from ..reader import DEFAULT_READERS
            readers = DEFAULT_READERS

        if isinstance(fstring, six.text_type):
            raise ReaderError('from_string expects a byte string, not a unicode string')

        for reader in readers:
            # Skip reader if we don't think it can read file
            if not reader.detect(fstring, fname=fname):
                continue
            try:
                d = reader.readstring(fstring)
                log.debug('Parsed document with %s' % reader.__class__.__name__)
                return d
            except ReaderError:
                pass
        raise ReaderError('Unable to read document')

    @property
    def elements(self):
        """
        A list of all the elements in this document. All elements subclass from :class:`~batterydataextractor.doc.element.BaseElement`,
        and represent things such as paragraphs or tables, and can be found in :mod:`batterydataextractor.doc.figure`,
        :mod:`batterydataextractor.doc.table_new`, and :mod:`batterydataextractor.doc.text`.
        """
        return self._elements

#     # TODO: memoized_property? Why doc.records.serialize() parse many times?
    @property
    def records(self):
        """
        All records found in this Document, as a list of :class:`~batterydataextractor.model.base.BaseModel`.
        """
        log.debug("Getting chemical records")
        records = ModelList()  # Final list of records -- output
        head_def_record = None  # Most recent record from a heading, title or short paragraph
        head_def_record_i = None # Element index of head_def_record
        last_product_record = None
        title_record = None # Records found in the title

        # Main loop, over all elements in the document
        for i, el in enumerate(self.elements):
            log.debug("Element %d, type %s" %(i, str(type(el))))
            last_id_record = None

            # FORWARD INTERDEPENDENCY RESOLUTION -- Updated model parsers to reflect defined entities
            # 1. Find any defined entities in the element e.g. "Curie Temperature, Tc"
            # 2. Update the relevant models
            # TODO: recover definition
            # element_definitions = el.definitions
            # for model in el.models:
            #     model.update(element_definitions)

            el_records = el.records
            # Save the title compound
            if isinstance(el, Title):
                if len(el_records) == 1 and isinstance(el_records[0], Compound) and el_records[0].is_id_only:
                    title_record = el_records[0]

            # Reset head_def_record unless consecutive heading with no records
            if isinstance(el, Heading) and head_def_record is not None:
                if not (i == head_def_record_i + 1 and len(el.records) == 0):
                    head_def_record = None
                    head_def_record_i = None

            # Paragraph with single sentence with single ID record considered a head_def_record
            if isinstance(el, Paragraph) and len(el.sentences) == 1:
                if len(el_records) == 1 and isinstance(el_records[0], Compound) and el_records[0].is_id_only:
                    head_def_record = el_records[0]
                    head_def_record_i = i

            # Paragraph with multiple sentences
            # We assume that if the first sentence of a paragraph contains only 1 ID Record, we can treat it as a header definition record, unless directly proceeding a header def record
            elif isinstance(el, Paragraph) and len(el.sentences) > 0:
                if not (isinstance(self.elements[i - 1], Heading) and head_def_record_i == i - 1):
                    first_sent_records = el.sentences[0].records
                    if len(first_sent_records) == 1 and isinstance(first_sent_records[0], Compound) and first_sent_records[0].is_id_only:
                        sent_record = first_sent_records[0]
                        if sent_record.names:
                            head_def_record = sent_record
                            head_def_record_i = i

            #: BACKWARD INTERDEPENDENCY RESOLUTION BEGINS HERE
            for record in el_records:
                if isinstance(record, Compound):
                    # Keep track of the most recent compound record with labels
                    # Heading records with compound ID's
                    if isinstance(el, Heading) and record.names:
                        head_def_record = record
                        head_def_record_i = i
                        # If 2 consecutive headings with compound ID, merge in from previous
                        if i > 0 and isinstance(self.elements[i - 1], Heading):
                            prev = self.elements[i - 1]
                            if (len(el.records) == 1 and record.is_id_only and len(prev.records) == 1 and
                                isinstance(prev.records[0], Compound) and prev.records[0].is_id_only and
                                    not (record.names and prev.records[0].names)):
                                record.names.extend(prev.records[0].names)

                # Unidentified records -- those without compound names or labels
                if record.is_unidentified:
                    if hasattr(record, 'compound'):
                        # We have property values but no names or labels... try merge those from previous records
                        if isinstance(el, Paragraph) and (head_def_record or last_product_record or last_id_record or title_record):
                            # head_def_record from heading takes priority if the heading directly precedes the paragraph ( NOPE: or the last_id_record has no name)
                            if head_def_record_i and head_def_record_i + 1 == i: # or (last_id_record and not last_id_record.names)):
                                if head_def_record:
                                    record.compound = head_def_record
                                elif last_id_record:
                                    record.compound = last_id_record
                                elif last_product_record:
                                    record.compound = last_product_record
                                elif title_record:
                                    record.compound = title_record
                            else:
                                if last_id_record:
                                    record.compound = last_id_record
                                elif head_def_record:
                                    record.compound = head_def_record
                                elif last_product_record:
                                    record.compound = last_product_record
                                elif title_record:
                                    record.compound = title_record
                        else:
                            pass

                if record not in records:
                    log.debug(record.serialize())
                    records.append(record)

        # clean up records
        cleaned_records = ModelList()
        for record in records:
            if record.required_fulfilled and record not in cleaned_records:
                if (self.models and type(record) in self.models) or not self.models:
                    cleaned_records.append(record)

        # Reset updatables
        for el in self.elements:
            for model in el.models:
                model.reset_updatables()

        return cleaned_records

    def get_element_with_id(self, id):
        """
        Get element with the specified ID. If one is not found, None is returned.
        :param id: Identifier to search for.
        :returns: Element with specified ID
        :rtype: BaseElement or None
        """
        """Return the element with the specified ID."""
        # Should we maintain a hashmap of ids to make this more efficient? Probably overkill.
        # TODO: Elements can contain nested elements (captions, footnotes, table cells, etc.)
        return next((el for el in self.elements if el.id == id), None)

    @property
    def citations(self):
        """
        A list of all :class:`~batterydataextractor.doc.text.Citation` elements in this Document.
        """
        return [el for el in self.elements if isinstance(el, Citation)]

    @property
    def footnotes(self):
        """
        A list of all :class:`~batterydataextractor.doc.text.Footnote` elements in this Document.
        .. note::
            Elements (e.g. Tables) can contain nested Footnotes which are not taken into account.
        """
        # TODO: Elements (e.g. Tables) can contain nested Footnotes
        return [el for el in self.elements if isinstance(el, Footnote)]

    @property
    def titles(self):
        """
        A list of all :class:`~batterydataextractor.doc.text.Title` elements in this Document.
        """
        return [el for el in self.elements if isinstance(el, Title)]

    @property
    def headings(self):
        """
        A list of all :class:`~batterydataextractor.doc.text.Heading` elements in this Document.
        """
        return [el for el in self.elements if isinstance(el, Heading)]

    @property
    def paragraphs(self):
        """
        A list of all :class:`~batterydataextractor.doc.text.Paragraph` elements in this Document.
        """
        return [el for el in self.elements if isinstance(el, Paragraph)]

    @property
    def captions(self):
        """
        A list of all :class:`~batterydataextractor.doc.text.Caption` elements in this Document.
        """
        return [el for el in self.elements if isinstance(el, Caption)]

    @property
    def captioned_elements(self):
        """
        A list of all :class:`~batterydataextractor.doc.element.CaptionedElement` elements in this Document.
        """
        return [el for el in self.elements if isinstance(el, CaptionedElement)]

    @property
    def metadata(self):
        """Return metadata information
        """
        return [el for el in self.elements if isinstance(el, MetaData)]

    @property
    def abbreviation_definitions(self):
        """
        A list of all abbreviation definitions in this Document. Each abbreviation is in the form
        (:class:`str` abbreviation, :class:`str` long form of abbreviation, :class:`str` ner_tag)
        """
        return [ab for el in self.elements for ab in el.abbreviation_definitions]

    @property
    def ner_tags(self):
        """
        A list of all Named Entity Recognition tags in this Document.
        If a word was found not to be a named entity, the named entity tag is None,
        and if it was found to be a named entity, it can have either a tag of 'B-CM' for a beginning of a
        mention of a chemical or 'I-CM' for the continuation of a mention.
        """
        return [n for el in self.elements for n in el.ner_tags]

    @property
    def cems(self):
        """
        A list of all Chemical Entity Mentions in this document as :class:`~batterydataextractor.doc.text.Span`
        """
        return list(set([n for el in self.elements for n in el.cems]))

    def serialize(self):
        """
        Convert Document to Python dictionary. The dictionary will always contain the key 'type', which will be 'document',
        and the key 'elements', which contains a dictionary representation of each of the elements of the document.
        """
        # Serialize fields to a dict
        elements = []
        for element in self.elements:
            elements.append(element.serialize())
        data = {'type': 'document', 'elements': elements}
        return data

    def to_json(self, *args, **kwargs):
        """Convert Document to JSON string. The content of the JSON will be equivalent
        to that of :meth:`serialize`.
        The document itself will be under the key 'elements',
        and there will also be the key 'type', which will always be 'document'.
        Any arguments for :func:`json.dumps` can be passed into this function."""
        return json.dumps(self.serialize(), *args, **kwargs)

    def _repr_html_(self):
        html_lines = ['<div class="cde-document">']
        for element in self.elements:
            html_lines.append(element._repr_html_())
        html_lines.append('</div>')
        return '\n'.join(html_lines)

    def to_database(self, file_name, file_type='json'):
        """
        Save the document to a file in the database.

        :param file_name: The name of the file to save the document to.
        :param file_type: The type of file to save the document to.
        """
        if file_type == 'json':
            with open(file_name, 'a', encoding='utf-8') as f:
                json.dump(self.serialize(), f)
                f.write('\n')
        elif file_type == 'txt':
            with open(file_name, 'a', encoding='utf-8') as f:
                f.write(json.dumps(self.serialize()))
                f.write('\n')
        elif file_type == 'csv':
            data = self.serialize()
            dic = data[list(data.keys())[0]]
            dic['model_type'] = list(dic.keys())[0]
            csv_columns = dic.keys()
            with open(file_name, 'a', encoding='utf-8') as csvfile:
                writer = csv.DictWriter(csvfile, fieldnames=csv_columns)
                writer.writeheader()
                writer.writerow(dic)
        else:
            raise ValueError('Unknown file type. Please use json, txt, or csv.')



import logging
import unittest
import os

from batterydataextractor.doc.document import Document
from batterydataextractor.doc.text import Paragraph, Title, Heading, Caption, Footnote
# from batterydataextractor.config import Config
# from batterydataextractor.model import Compound, NmrSpectrum, IrSpectrum, UvvisSpectrum, MeltingPoint, GlassTransition
from batterydataextractor.nlp import *

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class TestText(unittest.TestCase):
    """Simple Text instantiation tests."""

    # @staticmethod
    # def get_config():
    #     test_config_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'test_config.yml')
    #     return Config(test_config_path)

    # def test_parsers_set_from_config(self):
    #     d = Document(Title('Test'), Paragraph('Test'), Footnote('Test'), Heading('Test'), Caption('Test'), config=self.get_config())
    #     title_p = d.titles[0].parsers
    #     heading_p = d.headings[0].parsers
    #     paragraph_p = d.paragraphs[0].parsers
    #     footnotes_p = d.footnotes[0].parsers
    #     caption_p = d.captions[0].parsers

    #     for parser in paragraph_p:
    #         self.assertIn(type(parser), [CompoundParser, ChemicalLabelParser, NmrParser])
    #     self.assertEqual(type(title_p[0]), NmrParser)
    #     self.assertEqual(type(footnotes_p[0]), ChemicalLabelParser)
    #     self.assertEqual(type(heading_p[0]), NmrParser)
    #     self.assertEqual(type(caption_p[0]), NmrParser)

    # def test_default_models(self):
    #     # Add some sort of test for ContextParser type stuff?
    #     d = Document(Title('Test'), Paragraph('Test'), Footnote('Test'), Heading('Test'), Caption('Test'))
    #     title_m = d.titles[0].models
    #     heading_m = d.headings[0].models
    #     paragraph_m = d.paragraphs[0].models
    #     footnotes_m = d.footnotes[0].models
    #     captions_m = d.footnotes[0].models
    #     for model in paragraph_m:
    #         self.assertIn(model, [Compound, NmrSpectrum, IrSpectrum, UvvisSpectrum, MeltingPoint, GlassTransition])
    #     self.assertEqual(title_m, [Compound])
    #     self.assertEqual(heading_m, [Compound])
    #     self.assertEqual(captions_m, [Compound])

    # def test_set_cem(self):
    #     d = Document(Title('Test'), config=self.get_config())
    #     title = d.titles[0]
    #     self.assertEqual(type(title.pos_tagger), CrfPosTagger)
    #     self.assertEqual(type(title.ner_tagger), CiDictCemTagger)
    #     self.assertEqual(type(title.lexicon), Lexicon)
    #     self.assertEqual(type(title.sentence_tokenizer), SentenceTokenizer)
    #     self.assertEqual(type(title.word_tokenizer), WordTokenizer)

    def test_no_cem_set(self):
        d = Document(Title('Test'))
        title = d.titles[0]
        self.assertEqual(type(title.pos_tagger), BertTagger)
        self.assertEqual(type(title.ner_tagger), CemTagger)
        self.assertEqual(type(title.lexicon), ChemLexicon)
        self.assertEqual(type(title.sentence_tokenizer), ChemSentenceTokenizer)
        self.assertEqual(type(title.word_tokenizer), ChemWordTokenizer)



import unittest
from batterydataextractor.nlp import BertTagger
from batterydataextractor.doc import Text


class TestBertTagger(unittest.TestCase):
    """Test BertTagger."""

    @classmethod
    def setUpClass(cls):
        cls.t = BertTagger()

    def test_tag_simple(self):
        """Test the ChemApPosTagger  on a simple sentence."""
        self.assertEqual(
            [('And', 'CC'), ('now', 'RB'), ('for', 'IN'), ('something', 'NN'), ('completely', 'RB'), ('different', 'JJ')],
            self.t.tag(['And', 'now', 'for', 'something', 'completely', 'different'])
        )

    def test_text_sentence(self):
        """Test tagging through the Text and Sentence API."""
        t = Text('Now for something completely different')
        self.assertEqual(
            [[(u'Now', u'RB'), (u'for', u'IN'), (u'something', u'NN'), (u'completely', u'RB'), (u'different', u'JJ')]],
            t.pos_tagged_tokens
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
batterydataextractor.scrape.rsc

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
The Royal Society of Chemistry (RSC) web scraper. Please get the permission from RSC before web-scraping.
"""
import re
import requests
from time import sleep
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from bs4 import BeautifulSoup
from .base import BaseWebScraper


class RSCWebScraper(BaseWebScraper):
    """
    RSC web-scraper
    """
    def __init__(self, url=None, max_wait_time=30, driver=None):
        """
        :param url: RSC search query url.
        :param max_wait_time: maximum waiting time for the scraper.
        :param driver: the Selenium driver (default: Chrome driver).
        """
        super().__init__()
        self.url = url
        self.max_wait_time = max_wait_time
        self.driver = driver

    def get_doi(self, query, page):
        """
        Get a list of dois from query massages and the exact page.
        :param query: the query text (e.g. battery materials)
        :param page: the number of page
        :return: a list of dois of the relevant query text and page.
        """
        if self.driver is None:
            driver = webdriver.Chrome()
        else:
            driver = self.driver

        if self.url is None:
            url = "http://pubs.rsc.org/en/results?searchtext="
            url = url + query
        else:
            url = self.url
        driver.get(url)
        wait = WebDriverWait(driver, self.max_wait_time)
        # To make sure we don't overload the server
        sleep(1)
        next_button = wait.until(EC.visibility_of_all_elements_located((By.CSS_SELECTOR, "a[class^=paging__btn]")))[1]
        page_string = """document.querySelectorAll("a[class^=paging__btn]")[1].setAttribute("data-pageno", \""""\
                      + str(page) + """\")"""
        driver.execute_script(page_string)
        next_button.click()
        sleep(1)
        _ = wait.until(EC.visibility_of_element_located((By.CSS_SELECTOR, '.capsule.capsule--article')))
        doi_lists = driver.find_elements(By.PARTIAL_LINK_TEXT, 'https://doi.org')
        dois = [doi.text for doi in doi_lists]
        return dois

    def download_doi(self, doi, file_location):
        """
        Download the html paper of the doi
        :param doi: doi of the paper
        :param file_location: the saving location
        :return:
        """
        doi = doi.split("org/")[-1]
        r = requests.get('http://doi.org/' + doi, headers={'User-Agent': 'Mozilla/5.0'})
        result = re.findall(r'https://pubs.rsc.org/en/content/articlehtml/.*?"', r.text)
        url = result[0][:-1]
        web_content = requests.get(url, headers={"User-Agent": "Mozilla/5.0"}).content

        result = self.get_rsc_abstract(web_content)
        exact_date = result['date'].split("/")
        doi = result['doi'].replace("/", "_")
        if len(exact_date) == 3:
            name = exact_date[0] + exact_date[1] + exact_date[2] + '_' + doi
        else:
            name = result['online_date'].replace("/", "") + '_' + doi

        with open(file_location + name + '.html', 'wb') as f:
            f.write(web_content)
        return

    @staticmethod
    def get_rsc_abstract(web_content):
        """
        Get the metadata and abstract from a rsc html file
        :param web_content: rsc html content
        :return: a dictionary of metadata and abstract
        """
        soup = BeautifulSoup(web_content, features="html.parser")
        for i in soup.find_all("meta"):
            if i.has_attr("name"):
                if i['name'] == "DC.Identifier":
                    doi = i['content']
            if i.has_attr("name"):
                if i["name"] == "DC.Date":
                    date = i['content']
            if i.has_attr("name"):
                if i["name"] == "DC.title":
                    title = i['content']
            if i.has_attr("name"):
                if i["name"] == "citation_journal_title":
                    journal = i['content']
            if i.has_attr("name"):
                if i["name"] == "citation_online_date":
                    online_date = i['content']
        abstract = soup.find_all("p", attrs={"class": "abstract"})[0].get_text()
        return {"title": title, "doi": doi, "date": date, "journal": journal, "online_date": online_date,
                "abstract": abstract}


#: Map image URL components to unicode characters.
RSC_IMG_CHARS = {
    '2041': '^',              # caret
    '224a': '\u2248',         # almost equal
    'e001': '=',              # equals
    'e002': '\u2261',         # equivalent
    'e003': '\u2263',         # strictly equivalent
    'e006': '=',              # equals
    'e007': '\u2261',         # equivalent
    'e009': '>',              # greater than
    'e00a': '<',              # less than
    'e00c': '\u269f',         # three lines converging left
    'e00d': '\u269e',         # three lines converging right
    'e010': '\u250c',         # box down and right
    'e011': '\u2510',         # box down and left
    'e012': '\u2514',         # box up and right
    'e013': '\u2518',         # box up and left
    'e038': '\u2b21',         # white hexagon
    'e059': '\u25cd',         # ?
    'e05a': '\u25cd',         # ?
    'e069': '\u25a9',         # square with diagonal crosshatch fill
    'e077': '\u2b13',         # square with bottom half black
    'e082': '\u2b18',         # diamond with top half black
    'e083': '\u2b19',         # diamond with bottom half black
    'e084': '\u27d0',         # white diamond with centred do
    'e090': '\u2504',         # box drawings light triple dash horizontal (not exactly)
    'e091': '\u2504',         # box drawings light triple dash horizontal
    'e0a2': '\u03b3\u0307',   # small gamma with dot
    'e0b3': '\u03bc\u0342',   # small mu with circumflex
    'e0b7': '\u03c1\u0342',   # small rho with circumflex
    'e0c2': '\u03b1\u0305',   # small alpha with macron
    'e0c3': '\u03b2\u0305',   # small beta with macron
    'e0c5': '\u03b4\u0305',   # small delta with macron
    'e0c6': '\u03b5\u0305',   # small epsilon with macron
    'e0ce': 'v\u0305',        # small v with macron
    'e0c9': '\u03b8\u0305',   # small theta with macron
    'e0cb': '\u03ba\u0305',   # small kappa with macron
    'e0cc': '\u03bb\u0305',   # small lambda with macron
    'e0cd': '\u03bc\u0305',   # small mu with macron
    'e0d1': '\u03c1\u0305',   # small rho with macron
    'e0d4': '\u03c4\u0305',   # small tau with macron
    'e0d5': '\u03bd\u0305',   # small nu with macron
    'e0d6': '\u03d5\u0305',   # small phi with macron (stroked)
    'e0d7': '\u03c6\u0305',   # small phi with macron
    'e0d8': '\u03c7\u0305',   # small chi with macron
    'e0da': '\u03bd\u0305',   # small omega with macron
    'e0db': '\u03a6\u0303',   # capital phi with tilde
    'e0dd': '\u03b3\u0303',   # small lambda with tilde
    'e0de': '\u03b5\u0303',   # small epsilon with tilde
    'e0e0': '\u03bc\u0303',   # small mu with tilde
    'e0e1': 'v\u0303',        # small v with tilde
    'e0e4': '\u03c1\u0303',   # small rho with tilde
    'e0e7': '\u03b5\u20d7',   # small epsilon with rightwards arrow above
    'e0e9': '\u03bc\u20d7',   # small mu with rightwards arrow above
    'e0eb': '\u29b5',         # circle with horizontal bar
    'e0ec': '|',              # ? http://www.rsc.org/images/entities/char_e0ec.gif
    'e0ed': '|',              # ? http://www.rsc.org/images/entities/char_e0ed.gif
    'e0ee': '3/2',            # 3/2
    'e0f1': '\U0001d302',     # ?
    'e0f5': '\u03bd',         # small nu
    'e0f6': '\u27ff',         # long rightwards squiggle arrow
    'e100': '\u2506',         # box drawings light triple dash vertical
    'e103': '\u2605',         # Black Star
    'e107': '\u03b5\u0342',   # small epsilon with circumflex
    'e108': '\u03b7\u0342',   # small eta with circumflex
    'e109': '\u03ba\u0342',   # small kappa with circumflex
    'e10d': '\u03c3\u0303',   # small sigma with tilde
    'e110': '\u03b7\u0303',   # small eta with tilde
    'e112': '\U0001d4a2',     # script G
    'e113': '\U0001d219',     # ? greek vocal notation symbol-51
    'e116': '\u2933',         # wave arrow pointing directly right
    'e117': '\u2501',         # box drawings heavy horizontal
    'e11a': '\u03bb\u0342',   # small lambda with circumflex
    'e11b': '\u03c7\u0303',   # small chi with tilde
    'e11f': '5/2',            # 5/2
    'e120': '5/4',            # 5/4
    'e124': '\u2b22',         # black hexagon
    'e131': '\u03bd\u0303',   # small nu with tilde
    'e132': '\u0393\u0342',   # capital gamma with circumflex
    'e13d': '\u2b1f',         # black pentagon
    'e142': '\u210b',         # script capital H
    'e144': '\u2112',         # script capital L
    'e146': '\u2113',         # script small l
    'e170': '\U0001d544',     # double-struck capital M
    'e175': '\u211d',         # double-struck capital R
    'e177': '\U0001d54b',     # double-struck capital T
    'e17e': '\U0001D580',     # fraktur bold capital U
    'e18f': '\U0001d57d',     # fraktur bold capital R
    'e1c0': '\u2b21',         # white hexagon
    'e520': '\U0001d49c',     # script capital A
    'e523': '\U0001d49f',     # script capital D
    'e529': '\U0001d4a5',     # script capital J
    'e52d': '\U0001d4a9',     # script capital N
    'e52f': '\U0001d4ab',     # script capital P
    'e531': '\u211b',         # script capital R
    'e533': '\U0001d4af',     # script capital T
}


def rsc_html_whitespace(document):
    """ Remove whitespace in xml.text or xml.tails for all elements, if it is only whitespace """
    # selects all tags and checks if the text or tail are spaces
    for el in document.xpath('//*'):
        if el.tag == 'b':
            continue
        if str(el.text).isspace():
            el.text = ''
        if str(el.tail).isspace():
            el.tail = ''
        if el.text:
            el.text = el.text.replace('\n', ' ')
    return document


def join_rsc_table_captions(document):
    """Add wrapper tag around Tables and their respective captions
    Arguments:
        document {[type]} -- [description]
    """
    for el in document.xpath('//div[@class="table_caption"]'):
        next_el = el.getnext()
        if next_el.tag == 'div' and next_el.attrib['class'] == 'rtable__wrapper':
            caption_el = el
            table_el = next_el
            table_el.insert(0, caption_el)
    return document


def replace_rsc_img_chars(document):
    """Replace image characters with unicode equivalents."""
    image_re = re.compile('http://www.rsc.org/images/entities/(?:h[23]+_)?(?:[ib]+_)?char_([0-9a-f]{4})(?:_([0-9a-f]{4}))?\.gif')
    for img in document.xpath('.//img[starts-with(@src, "http://www.rsc.org/images/entities/")]'):
        m = image_re.match(img.get('src'))
        if m:
            u1, u2 = m.group(1), m.group(2)
            if not u2 and u1 in RSC_IMG_CHARS:
                rep = RSC_IMG_CHARS[u1]
            else:
                rep = ('\\u%s' % u1).encode('ascii').decode('unicode-escape')
                if u2:
                    rep += ('\\u%s' % u2).encode('ascii').decode('unicode-escape')
            if img.tail is not None:
                rep += img.tail  # Make sure we don't remove any tail text
            parent = img.getparent()
            if parent is not None:
                previous = img.getprevious()
                if previous is not None:
                    previous.tail = (previous.tail or '') + rep
                else:
                    parent.text = (parent.text or '') + rep
                parent.remove(img)
    return document


def rsc_clean_abstract(document):
    """ Remove <h2> in abstract"""
    # selects all tags and checks if the text or tail are spaces
    for el in document.xpath('//div[@class="abstract"] | //p[@class="abstract"]'):
        next_els = el.getchildren()
        for next_el in next_els:
            if next_el.tag == 'h2':
                parent = next_el.getparent()
                if parent is None:
                    continue
                if next_el.tail:
                    previous = next_el.getprevious()
                    if previous is None:
                        parent.text = (parent.text or '') + next_el.tail
                    else:
                        previous.tail = (previous.tail or '') + next_el.tail
                parent.remove(next_el)
    return document



# -*- coding: utf-8 -*-
"""
batterydataextractor.scrape.__init__

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Web-scrapers from publishers including RSC, Elsevier, and Springer
"""
from .base import BaseWebScraper
from .rsc import RSCWebScraper
from .elsevier import ElsevierWebScraper
from .springer import SpringerMetaWebScraper, SpringerTDMWebScraper



# -*- coding: utf-8 -*-
"""
test_reader_rsc
~~~~~~~~~~~~~~~
Test RSC reader.
"""
import io
import logging
import os
import unittest

from batterydataextractor.doc import Document
from batterydataextractor.reader import RscHtmlReader


logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class TestRscHtmlReader(unittest.TestCase):

    maxDiff = None

    def test_detect(self):
        """Test RscHtmlReader can detect an RSC document."""
        r = RscHtmlReader()
        fname = 'rsc_test1.html'
        f = io.open(os.path.join(os.path.dirname(__file__), 'testpapers', fname), 'rb')
        content = f.read()
        self.assertEqual(r.detect(content, fname=fname), True)

    def test_direct_usage(self):
        """Test RscHtmlReader used directly to parse file."""
        r = RscHtmlReader()
        fname = 'rsc_test1.html'
        f = io.open(os.path.join(os.path.dirname(__file__), 'testpapers', fname), 'rb')
        content = f.read()
        d = r.readstring(content)
        self.assertEqual(len(d.elements), 95)

    def test_document_usage(self):
        """Test RscHtmlReader used via Document.from_file."""
        fname = 'rsc_test2.html'
        f = io.open(os.path.join(os.path.dirname(__file__), 'testpapers', fname), 'rb')
        d = Document.from_file(f, readers=[RscHtmlReader()])
        self.assertEqual(len(d.elements), 60)


if __name__ == '__main__':
    unittest.main()
